@@ -10,11 +10,11 @@ from __future__ import absolute_import
 
 import fcntl
 import io
+import logging
 import mmap
 import os
 
 from contextlib import closing
-from contextlib import contextmanager
 
 from . import util
 from . import errors
@@ -27,6 +27,8 @@ BUFFERSIZE = 1024 * 1024
 # sufficient for doing direct I/O.
 BLOCKSIZE = 512
 
+log = logging.getLogger("directio")
+
 
 class EOF(Exception):
     """ Raised when no more data is available and size was not specifed """
@@ -44,6 +46,8 @@ class Operation(object):
             self._buffersize = buffersize
         self._done = 0
         self._active = True
+        self._clock = util.Clock()
+        self._clock.start("total")
 
     @property
     def size(self):
@@ -71,10 +75,13 @@ class Operation(object):
         try:
             self._run()
         finally:
-            self._active = False
+            self.close()
 
     def close(self):
-        self._active = False
+        if self._active:
+            self._active = False
+            self._clock.stop("total")
+            log.info("Operation stats: %s", self._clock)
 
     def __repr__(self):
         return ("<{self.__class__.__name__} path={self._path!r} "
@@ -99,12 +106,14 @@ class Send(Operation):
 
     def _run(self):
         for chunk in self:
+            self._clock.start("write")
             self._dst.write(chunk)
+            elapsed = self._clock.stop("write")
+            log.debug("Wrote %d bytes in %.3f seconds", len(chunk), elapsed)
 
     def __iter__(self):
-        with io.FileIO(self._path, "r") as src, \
-                aligned_buffer(self._buffersize) as buf:
-            enable_directio(src.fileno())
+        with open(self._path, "r") as src, \
+                closing(aligned_buffer(self._buffersize)) as buf:
             try:
                 if self._offset:
                     skip = self._seek_to_first_block(src)
@@ -125,7 +134,10 @@ class Send(Operation):
                 raise EOF
             raise errors.PartialContent(self.size, self.done)
 
+        self._clock.start("read")
         count = util.uninterruptible(src.readinto, buf)
+        elapsed = self._clock.stop("read")
+        log.debug("Read %d bytes in %.3f seconds", count, elapsed)
         if count == 0:
             if self._size is None:
                 raise EOF
@@ -147,21 +159,24 @@ class Receive(Operation):
         self._src = src
 
     def _run(self):
-        with io.FileIO(self._path, "r+") as dst, \
-                aligned_buffer(self._buffersize) as buf:
+        with open(self._path, "r+") as dst, \
+                closing(aligned_buffer(self._buffersize)) as buf:
             try:
                 if self._offset:
                     remaining = self._seek_before_first_block(dst)
                     if remaining:
+                        disable_directio(dst.fileno())
                         self._receive_chunk(dst, buf, remaining)
-                enable_directio(dst.fileno())
+                        enable_directio(dst.fileno())
                 while self._todo:
                     count = min(self._todo, self._buffersize)
                     self._receive_chunk(dst, buf, count)
             except EOF:
                 pass
             finally:
+                self._clock.start("sync")
                 os.fsync(dst.fileno())
+                self._clock.stop("sync")
 
     def _seek_before_first_block(self, dst):
         dst.seek(self._offset)
@@ -174,7 +189,10 @@ class Receive(Operation):
         buf.seek(0)
         toread = count
         while toread:
+            self._clock.start("read")
             chunk = self._src.read(toread)
+            elapsed = self._clock.stop("read")
+            log.debug("Read %d bytes in %.3f seconds", len(chunk), elapsed)
             if chunk == "":
                 break
             buf.write(chunk)
@@ -187,7 +205,11 @@ class Receive(Operation):
             wbuf = buffer(buf, offset, size)
             if size % BLOCKSIZE:
                 disable_directio(dst.fileno())
-            towrite -= util.uninterruptible(dst.write, wbuf)
+            self._clock.start("write")
+            written = util.uninterruptible(dst.write, wbuf)
+            elapsed = self._clock.stop("write")
+            log.debug("Wrote %d bytes in %.3f seconds", written, elapsed)
+            towrite -= written
 
         self._done += buf.tell()
         if buf.tell() < count:
@@ -196,12 +218,37 @@ class Receive(Operation):
             raise errors.PartialContent(self.size, self.done)
 
 
+def open(path, mode, direct=True):
+    """
+    Open a file for direct I/O.
+
+    Writing or reading from the file requires an aligned buffer. Only
+    readinto() can be used to read from the file.
+
+    If direct is False, open the file for buffered I/O. You can enable direct
+    I/O later using enable_directio().
+    """
+    if mode == "r":
+        flags = os.O_RDONLY
+    elif mode == "w":
+        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
+    elif mode == "r+":
+        flags = os.O_RDWR
+    else:
+        raise ValueError("Unsupported mode %r" % mode)
+
+    if direct:
+        flags |= os.O_DIRECT
+
+    fd = os.open(path, flags)
+    return io.FileIO(fd, mode, closefd=True)
+
+
 def round_up(n, size=BLOCKSIZE):
     n = n + size - 1
     return n - (n % size)
 
 
-@contextmanager
 def aligned_buffer(size):
     """
     Return buffer aligned to page size, which work for doing direct I/O.
@@ -213,9 +260,7 @@ def aligned_buffer(size):
     behavior in the parent or the child processes. This restriction does not
     apply to memory buffer created with MAP_SHARED. See open(2) for more info.
     """
-    buf = mmap.mmap(-1, size, mmap.MAP_SHARED)
-    with closing(buf):
-        yield buf
+    return mmap.mmap(-1, size, mmap.MAP_SHARED)
 
 
 def enable_directio(fd):
